library(sf)
library(raster)
library(stars)
library(tidyverse)
library(viridis)
library(geojsonsf)
library(rgdal)
library(tictoc) # For script performance monitoring
library(crayon) # For colouring console text
library(png)
library(RJSONIO)

# source functions
source("./load_gridnumbers.R")
source("./filter_distance.R")
source("./recommend_rank.R")
source("./recommend_metric.R")

# Logging variables
timing <- T
timeThreshold <- 1

# Projections
BNG <- "+proj=tmerc +lat_0=49 +lon_0=-2 +k=0.9996012717 +x_0=400000 +y_0=-100000 +ellps=airy +datum=OSGB36 +units=m +no_defs"
WGS84 <- "+proj=longlat +ellps=WGS84 +datum=WGS84 +no_defs"
WGS84_PseudoMercator <- "+proj=merc +a=6378137 +b=6378137 +lat_ts=0.0 +lon_0=0.0 +x_0=0.0 +y_0=0 +k=1.0 +units=m +nadgrids=@null +wktext  +no_defs"
crs <- list('27700'=BNG, '4326'=WGS84, '3857'= WGS84_PseudoMercator)

doTic <- function(message) {
  # Conditional performance timing - only plot if timing set to true
  # Allows us to switch off timing globally
  if (timing) {
    tic(message, quiet=FALSE, func.tic=function(tic, toc, msg) {
      sprintf("%s...", message)
    })
  }
}

doToc <- function() {
  if (timing) {
    toc(func.toc=function(tic, toc, msgOrig) {
      msg <- sprintf("Took: %s seconds (%s)", round(toc-tic, digits=2), msgOrig)
      if (toc-tic < timeThreshold) {
        blue(msg)
      } else {
        red(msg)
      }
    })
  }
}

createPngRaw <- function(ras, search_sf_epsg, epsg, style, min, max) {
  
  #doTic("createPng - reproject main raster")
  # Specify current projection
  #crs(ras) <- CRS(crs[['27700']])
  # Reproject so suitable for display in leaflet
  #ras <- projectRaster(ras, crs=crs[[epsg]])
  #doToc()
  
  doTic('createPng - colour')

  if (style == "rev_terrain") {
    cols <- rev(terrain.colors(255))
  }
  if (style == "blue_orange_red") {
    cols <- colorRampPalette(c("blue", "lightblue", "orange", "red"))(100)
  }
  if (style == "viridis") {
    cols <- viridis(100)
  }
  if (style == "viridis_truncated") {
    cols <- tail(viridis(140), 100)
  }
  if (style == "cividis") {
    cols <- cividis(100)
  }
  if (style == "reds") {
    cols <- colorRampPalette(c('#fff5f0','#fee0d2','#fcbba1','#fc9272','#fb6a4a','#ef3b2c','#cb181d','#a50f15','#67000d'))(100)
  }
  if (style == "yellow_red") {
    cols <- colorRampPalette(c('#ffffcc','#ffeda0','#fed976','#feb24c','#fd8d3c','#fc4e2a','#e31a1c','#bd0026','#800026'))(100)
  }
  
  
  # zlim=c(min,max), zlimcol="transparent"
  rasArray <- as.array(RGB(ras, alpha=TRUE, colNA="black", col=cols, zlim=c(min,max), zlimcol="transparent"))/255
  doToc()

  
  # Make black pixels transparent
  trans <- which((rasArray[,,1] == 0) & (rasArray[,,2] == 0) & (rasArray[,,3] == 0))

  #blah <- which((rasArray[,,1] != 0) | (rasArray[,,2] != 0) | (rasArray[,,3] != 0))

  rasArray[,,4][trans]  <- 0
  #rasArray[,,4][blah]  <- 40


  #print('trans')
  #str(trans)
  #print('trans end')
  
  doTic('create Raw Png')
  rawPng <- writePNG(rasArray)
  doToc()
  
  # Return raw png
  return(rawPng)
}

generateRiskJsonByNames <- function(body) {

  # This was created to test the heatmap idea and not currently used
  
  json <- fromJSON(body)
  lat <- as.numeric(json$lat)
  lon <- as.numeric(json$lon)
  names <- json$names
  radius <- as.numeric(json$rad)

  location = c(lon, lat) 
  distance = radius
  
  doTic('predicted probability distributions')
  predf <- map_chr(names, function(x) sprintf('./data/rf_models/rf_SDMs_%s_meanpred.grd', x))
  preds <- raster::stack(predf)
  names(preds) <- names
  doToc()
  
  doTic('variation layers')
  varf <- map_chr(names, function(x) sprintf('./data/rf_models/rf_SDMs_%s_quantilerange.grd', x))
  var <- raster::stack(varf)
  names(var) <- names
  doToc()
  
  doTic('crop species data - preds')
  crop_pred <- filter_distance(obj = preds,
                               location = location,
                               distance = distance,
                               method = 'buffer')
  doToc()
  
  doTic('crop species data - var')
  crop_var <- filter_distance(obj = var,
                              location = location,
                              distance = distance,
                              method = 'buffer')
  doToc()
  
  doTic('create the decide score')
  additive_score <- recommend_metric(prediction_raster = crop_pred,
                                     error_raster = crop_var,
                                     method = 'additive')$additive
  doToc()
  
  doTic('Get the metric aggregated across all specyes')
  aggregate_score <- recommend_rank(predict_err_raster = additive_score,
                                    method = 'additive')$error_metric
  doToc()
  
  doTic("Reproject main raster")
  # Specify current projection
  crs(aggregate_score) <- CRS(crs[['27700']])
  # Reproject so suitable for display in leaflet
  aggregate_score <- projectRaster(aggregate_score, crs=crs[['4326']])
  doToc()
  
  doTic("Make return array")
  xyz <- rasterToPoints(aggregate_score)
  doToc()
  
  #doTic("Reorder columns")
  #yzx <- xyz[, c(2, 1, 3)]
  #doToc()
  
  return(xyz)
}

getPreGeneratedImage <- function(lat, lon, radius, name, local, threshold, colour) {
  # This is for returning area of intereste from pre-generated image for a group of taxa

  location = c(lon, lat) 
  distance = radius

  file <- sprintf('./data/species_data/combined_decide_scores/variation_weighted_time_since_last_record/%s_DECIDE_Score_variation_weighted_by_time_since.grd', name)
  doTic(paste('get pre-calculated raster', file))
  pred <- raster(file)
  doToc()
  
  doTic('crop data')
  crop_pred <- filter_distance(obj = pred,
                               location = location,
                               distance = distance,
                               method = 'buffer')
  doToc()

  # If local flag is set then pass min/max values of local image
  # otherwise min/max of uncropped (for colour scaling)
  if (local) {
    min <- minValue(crop_pred)
    max <- maxValue(crop_pred)
  } else {
    min <- minValue(pred)
    max <- maxValue(pred)
  }
  
  #print('min')
  #print(min)
  #print('max')
  #print(max)
  
  #crop_pred[crop_pred < min+(max-min)*threshold] <- NA
  q <- quantile(crop_pred, probs = c(threshold))
  #print('quantiles')
  #print(q)
  crop_pred[crop_pred < q[[1]]] <- NA
  

  doTic("Make return image")
  pngImage <- createPngRaw(crop_pred, null, '3857', colour, min, max)
  doToc()
  
  return(readBin(pngImage,'raw', n=length(pngImage)))
}

generateImageByNames <- function(lat, lon, radius, names) {

  # This was made for demonstrating generation of score on the fly from different taxa
  
  location = c(lon, lat) 
  distance = radius
  
  doTic('predicted probability distributions')
  predf <- map_chr(names, function(x) sprintf('./data/rf_models/rf_SDMs_%s_meanpred.grd', x))
  preds <- raster::stack(predf)
  names(preds) <- names
  doToc()
  
  doTic('variation layers')
  varf <- map_chr(names, function(x) sprintf('./data/rf_models/rf_SDMs_%s_quantilerange.grd', x))
  var <- raster::stack(varf)
  names(var) <- names
  doToc()
  
  doTic('crop species data - preds')
  crop_pred <- filter_distance(obj = preds,
                               location = location,
                               distance = distance,
                               method = 'buffer')
  doToc()
  
  doTic('crop species data - var')
  crop_var <- filter_distance(obj = var,
                              location = location,
                              distance = distance,
                              method = 'buffer')
  doToc()
  
  doTic('create the decide score')
  additive_score <- recommend_metric(prediction_raster = crop_pred,
                                     error_raster = crop_var,
                                     method = 'additive')$additive
  doToc()
  
  doTic('Get the metric aggregated across all specyes')
  aggregate_score <- recommend_rank(predict_err_raster = additive_score,
                                    method = 'additive')$error_metric
  doToc()
  
  doTic("Make return image")
  pngImage <- createPngRaw(aggregate_score, null, '3857', 'viridis')
  doToc()
  
  return(readBin(pngImage,'raw', n=length(pngImage)))
}

generateImage <- function(lat, lon, radius) {
  
  location = c(lon, lat) 
  distance = radius

  doTic('species names')
  names <- unique(gsub(pattern = "rf_SDMs_|_meanpred.grd|_meanpred.gri", 
                      replacement = '',
                      x = list.files('./data/rf_models/', pattern = 'meanpred')))
  doToc()

  doTic('predicted probability distributions')
  preds <- raster::stack(list.files('./data/rf_models/', 
                                    pattern = 'meanpred.grd',
                                    full.names = TRUE))
  names(preds) <- names
  doToc()

  doTic('variation layers')
  var <- raster::stack(list.files('./data/rf_models/', 
                                  pattern = 'quantilerange.grd',
                                  full.names = TRUE))
  names(var) <- names
  doToc()

  doTic('crop species data - preds')
  crop_pred <- filter_distance(obj = preds,
                              location = location,
                              distance = distance,
                              method = 'buffer')
  doToc()

  doTic('crop species data - var')
  crop_var <- filter_distance(obj = var,
                              location = location,
                              distance = distance,
                              method = 'buffer')
  doToc()

  doTic('create the decide score')
  additive_score <- recommend_metric(prediction_raster = crop_pred,
                                    error_raster = crop_var,
                                    method = 'additive')$additive
  doToc()

  doTic('Get the metric aggregated across all specyes')
  aggregate_score <- recommend_rank(predict_err_raster = additive_score,
                                    method = 'additive')$error_metric
  doToc()

  doTic("Make return image")
  pngImage <- createPngRaw(aggregate_score, null, '3857', 'viridis')
  doToc()
  
  return(readBin(pngImage,'raw', n=length(pngImage)))
}

generateImagePost <- function(body) {
  
  #body = '{"lat": "51.602436","lon": "-1.110557", "rad": "5000", "names": ["t1", "t2"]}'
  #body = '{"lat": "51.602436","lon": "-1.110557", "rad": "5000", "names": ["Adscita_geryon","Adscita_statices","Anarta_melanopa","Anarta_myrtilli","Archiearis_parthenias"]}'
  json <- fromJSON(body)
  lat <- json$lat 
  lon <- json$lon
  names <- json$names
  rad <- json$rad
  return(generateImageByNames(as.numeric(lat), as.numeric(lon), as.numeric(rad), names))
}

accessGeoJsonOld <- function(lat, lon, radius, type) {
  
  # Wallingford
  # lat=51.602436
  # lon=-1.110557
  # radius=5000
  # type='greenspace' #prow, greenpsace, accesspoint, accessland
  
  location = c(lon, lat) 
  distance = radius
  
  # load a UK grid - currently 10km
  grid <- st_read('./data/grids/uk_grid_10km.shp')
  
  # find grid numbers that location + buffered zone covers
  grid_numbers <- load_gridnumbers(location = location,
                                   distance = distance,
                                   grid = grid)
  
  #prow, greenspace, accesspoint, accessland

  acc_loc <- lapply(c(1:length(grid_numbers)), FUN = function(n){
    files <- list.files(sprintf('./data/grids/%s_grid', type),
                        full.names = T,
                        pattern = paste0('_', grid_numbers[n], '.shp'))
    
    if (length(files) > 0) {
      access <- sf::st_read(files, quiet = TRUE)
      st_crs(access) <- 27700
    } else {
      access = NULL
    }
    return(access)
  })
  all_outs <- do.call(rbind, acc_loc)
  
  class(all_outs)[1] 
  
  # crop shapes to exact region of interest
  final_acc_loc <- filter_distance(all_outs,
                                  location = location,
                                  distance = distance,
                                  method = 'buffer')
  
  print(final_acc_loc)
  
  if (is.null(final_acc_loc)) {
    return ('')
  } else {
    trans_final_acc_loc <- st_transform(final_acc_loc, 4326)
    return(sf_geojson(trans_final_acc_loc))
  }
}

accessGeoJson <- function(lat, lon, radius, type) {
  
  # Wallingford
  # lat=51.602436
  # lon=-1.110557
  # radius=5000
  # type='greenspace' #prow, greenpsace, accesspoint, accessland
  
  location = c(lon, lat) 
  distance = radius
  
  # load a UK grid - currently 10km
  grid <- st_read('./data/grids/uk_grid_10km.shp')
  
  # find grid numbers that location + buffered zone covers
  grid_numbers <- load_gridnumbers(location = location,
                                   distance = distance,
                                   grid = grid)
  
  #prow, greenspace, accesspoint, accessland, nattrust, london
  
  if (type == 'prow') {
    folder = 'rowmaps_footpathbridleway/rowmaps_footpathbridleway/gridded_data_10km'
    suffix = 'prow_gridnumber_'
  } else if (type == 'greenspace') {
    folder = 'OS_greenspaces/OS Open Greenspace (ESRI Shape File) GB/data/gridded_greenspace_data_10km'
    suffix = 'grnspc_gridnumber_'
  } else if (type == 'accesspoint') {
    folder = 'OS_greenspaces/OS Open Greenspace (ESRI Shape File) GB/data/gridded_accesspoint_data_10km'
    suffix = 'accspnt_gridnumber_'
  } else if (type == 'accessland') {
    folder = 'CRoW_Act_2000_-_Access_Layer_(England)-shp/gridded_data_10km'
    suffix = 'access_land_gridnumber_'
  } else if (type == 'nattrust') {
    folder = 'national_trust/gridded_data_10km'
    suffix = 'national_trust_gridnumber_'
  } else if (type == 'london') {
    folder = 'greater-london-latest-free/london_gridded_data_10km'
    suffix = 'lond_paths_gridnumber_'
  }
  
  acc_loc <- lapply(c(1:length(grid_numbers)), FUN = function(n){
    files <- list.files(sprintf('./data/access/%s', folder),
                        full.names = T,
                        pattern = paste0('_', grid_numbers[n], '.shp'))
    
    if (length(files) > 0) {
      access <- sf::st_read(files, quiet = TRUE)
      st_crs(access) <- 27700
    } else {
      access = NULL
    }
    return(access)
  })
  all_outs <- do.call(rbind, acc_loc)
  
  class(all_outs)[1] 
  
  # crop shapes to exact region of interest
  final_acc_loc <- filter_distance(all_outs,
                                   location = location,
                                   distance = distance,
                                   method = 'buffer')
  
  if (is.null(final_acc_loc)) {
    return ('')
  } else {
    trans_final_acc_loc <- st_transform(final_acc_loc, 4326)
    return(sf_geojson(trans_final_acc_loc))
  }
}